// 2018-06-02 note: commented out ISR() in telemetry.cpp to avoid linker error with newer Arduino IDE.  I'm not using telemetry now anyway as my current transmitter is Spektrum.
// comment out following line (with //) for analogue joystick/button joystick version
#define ENCODER

// comment out following line (with //) for direct (parallel) 16x2 LCD, leave in for 16x2 LCD with I2C interface
#define I2C_LCD

#include <EEPROM.h>
#include "eeprom.h"
#ifdef ENCODER
  #include "encoder.h"
#else
  #include "joystick.h"
#endif
#ifdef I2C_LCD
  #include <Wire.h>
  #include <LiquidCrystal_I2C.h>
  LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE); // depending what type of I2C_LCD module you have, you may need to change the address (0x27) and the pin order.
#else
  #include <LiquidCrystal.h>
  LiquidCrystal lcd(12, 11, 7, 6, 5, 4); // initialize library to use specified pins
#endif
#include "cppm_out.h"
#include "button.h"
#include "telemetry.h"

float batteryVolts;
float telemVolts;
uint8_t menu; // which menu item is displayed (or 0 for 'main screens')
uint16_t remainingTime;
uint8_t mode; // motor running/stopped and joystick mode
uint8_t tempModelNum; // used when selecting model
uint8_t sweepMode;
uint8_t oldSweepMode;
int8_t posn; // pointer for editing different menus - reset to 0 when different menu selected
uint8_t buttonUp; // true if button is NOT currently pressed
uint8_t oldButtonState; // used when editing model name using button to go forward or back
uint8_t buttonMode; // decodes button presses and selects RIGHT, LEFT, KICK_UP, etc.
int16_t second; // one-second timer for buzzer and running countdown timer
boolean secondElapsed;

#ifdef ENCODER
  boolean encoderEdit; // true when editing a value, false when negotiating menus
  boolean encoderBeginEdit; // true on leading edge of edit (allows cursor to be switched on)
#endif


/*
void hexPrint(int i) {
  if (i < 16) {
    Serial.print(" ");
  }
  Serial.print(i, HEX);
}
*/

void setup(void) {
  boolean fiveButton; // true indicates a digital top/bottom/left/right joystick instead of an analogue one
  
  // Serial.begin(115200);

  pinMode(2, INPUT_PULLUP); // main 'single channel' push button input
  pinMode(8, OUTPUT); // buzzer
  pinMode(A3, INPUT_PULLUP); // for the five-button 'digital' joystick this is the 'right' button
  pinMode(A4, INPUT_PULLUP); // for the five-button 'digital' joystick this is the 'bottom' button  
  pinMode(A5, INPUT_PULLUP); // reset eeprom and calibrate joystick input.
  digitalWrite(8, LOW); // buzzer off
  

  lcd.begin(16, 2);
  
  if (!digitalRead(A5)) { // short pin A5 to ground to reset EEPROM at power on
    fiveButton = false; // default to an analogue joystick unless they press buttons on A3 or A4 during calibration
#ifndef ENCODER
    lcd.print(F("Cycle joystick."));
    lcd.setCursor(0, 1);
    lcd.print(F("Centre. Button."));
    eeprom::gd.joyMin[0] = 1023;
    eeprom::gd.joyMin[1] = 1023;
    eeprom::gd.joyMax[0] = 0;
    eeprom::gd.joyMax[1] = 0;

    while (digitalRead(2)) { // while main button not pressed sample joystick min max values (or buttons)
      int16_t a = analogRead(A0);
      if (a < eeprom::gd.joyMin[0])
        eeprom::gd.joyMin[0] = a;
      if (a > eeprom::gd.joyMax[0])
        eeprom::gd.joyMax[0] = a;
      
      a = analogRead(A1);
      if (a < eeprom::gd.joyMin[1])
        eeprom::gd.joyMin[1] = a;
      if (a > eeprom::gd.joyMax[1])
        eeprom::gd.joyMax[1] = a;

      if (!digitalRead(A3) || !digitalRead(A4)) // they must have built the digital joystick version
        fiveButton = true;
    }
    // when button is pressed sample centre values and reset eeprom
    eeprom::gd.joyCentre[0] = analogRead(A0);
    eeprom::gd.joyCentre[1] = analogRead(A1);
    eeprom::gd.fiveButton = fiveButton ? 1 : 0;
#endif

    eeprom::gd.modelNum = 1;
    eeprom::gd.batteryScale = 160;
    eeprom::gd.lowBatteryAlarm = 71;
    eeprom::putGD();
    
    eeprom::md.modelName[0] = 'U';
    eeprom::md.modelName[1] = 'n';
    eeprom::md.modelName[2] = 'u';
    eeprom::md.modelName[3] = 's';
    eeprom::md.modelName[4] = 'e';
    eeprom::md.modelName[5] = 'd';
    eeprom::md.modelName[6] = ' ';
    eeprom::md.modelName[7] = ' ';
    eeprom::md.modelName[8] = ' ';
    eeprom::md.modelName[9] = ' ';
    eeprom::md.modelName[10] = ' ';
    eeprom::md.modelName[11] = ' ';
    eeprom::md.modelName[12] = ' ';
    eeprom::md.modelName[13] = '\0';
    eeprom::md.rudderCentre[0] = 1500;
    eeprom::md.rudderCentre[1] = 1500;
    eeprom::md.rudderRight[0] = 2000;
    eeprom::md.rudderRight[1] = 1750;
    eeprom::md.rudderLeft[0] = 1000;
    eeprom::md.rudderLeft[1] = 1250;
    eeprom::md.elevatorCentre[0] = 1500;
    eeprom::md.elevatorCentre[1] = 1500;
    eeprom::md.elevatorUp[0] = 2000;
    eeprom::md.elevatorUp[1] = 1750;
    eeprom::md.elevatorDown[0] = 1000;
    eeprom::md.elevatorDown[1] = 1250;
    eeprom::md.motorOff = 1000;
    eeprom::md.motorOn = 1750;
    eeprom::md.motorSoftStart = 1000;
    eeprom::md.buttonSpeed = 250;
    eeprom::md.quickBlipEnabled = 1;
    eeprom::md.motorRunTime = 300;
    eeprom::md.Arange = 0; // 0 default to using A1 for model battery monitoring
    eeprom::md.Aalarm = 44; // 4.4V (reasonable value for A1 alarm 4-cell NiMH)

    for (uint8_t i = 1; i <= 27; i++)
      eeprom::putMD(i);  
  }
  
  eeprom::getGD();
  fiveButton = (eeprom::gd.fiveButton != 0);
  eeprom::getMD(eeprom::gd.modelNum);
  cppm_out::set_uS(0, eeprom::md.rudderCentre[0]); // channel 1 is rudder
  cppm_out::set_uS(1, eeprom::md.elevatorCentre[0]); // channel 2 is elevator
  cppm_out::set_uS(2, eeprom::md.motorOff); // channel 3 is motor

#ifdef ENCODER
  encoder::start();
  encoderEdit = false;
  encoderBeginEdit = false;
#else
  joystick::start(fiveButton);
#endif
  
  button::start();
  pinMode(A2, INPUT); // battery monitor via 10K 22K divider to give 16-volt input range
  batteryVolts = (float)eeprom::gd.batteryScale * analogRead(A2) / 10240; // initialize as reading is smoothed
  cppm_out::start();
  telemetry::start();
  menu = 0;
  mode = 0; 
  buttonMode = 0;
  second = 1000;
}

void runTimer(boolean timerRunning) {
  if (timerRunning) {
    if (remainingTime) --remainingTime;
    if (remainingTime == 0) {
      if (mode < 3) // in normal button mode, stop motor
        mode = 0;
      else // in joystick mode, pause motor (but it can't be restarted when remainingTime is zero)
        mode = 4;
    }
  }
}

void displayTime(uint16_t seconds) {
#ifdef ENCODER
  if (menu == 1) // display time on bottom row when editing
    lcd.setCursor(11, 1);
  else {
    if (telemetry::error) return;  
    lcd.setCursor(11, 0);
  }
#else
  if (telemetry::error) return;
  lcd.setCursor(11, 0);
#endif
  if (seconds < 600)
    lcd.print F(" ");
  
  lcd.print(seconds / 60);
  lcd.print F(":");

  seconds %= 60;
  if (seconds < 10)
    lcd.print F("0");
  
  lcd.print(seconds);
}

void displayTimer() {
  displayTime(remainingTime);
}

void topLine(const __FlashStringHelper *top) {
  lcd.setCursor(0, 0);
  lcd.print(top);  
}

void clearBottomLine() {
  lcd.setCursor(0, 1);
  lcd.print(F("                "));
}

void refresh(const __FlashStringHelper *top, const __FlashStringHelper *bottom, uint16_t mS) {
  static int16_t bottomLineCycler = 0;

  if (telemetry::error) {
    lcd.setCursor(0, 0);
    lcd.print(telemetry::errMsg);
    // Serial.print("#");
    // Serial.print(telemetry::errMsg);
    // Serial.print("#");
  } else {
    topLine(top);
  }
  
  bottomLineCycler -= mS;
  if (bottomLineCycler <= 0)
    bottomLineCycler += 4500;

  lcd.setCursor(0, 1);
  if (bottomLineCycler >= 3000) { // display battery voltage
    lcd.print(F("Battery "));
    if (batteryVolts <= 0.1 * eeprom::gd.lowBatteryAlarm) {
      lcd.print(F("low"));
    }
    else {
      lcd.print(F("   "));
    }
    if (batteryVolts < 10.0)
      lcd.print(F(" "));

    lcd.print((int)batteryVolts);
    lcd.print(F("."));
    float f = (batteryVolts - (int)batteryVolts);
    lcd.print((int)(10 * f));
    lcd.print(F("V"));
  }
  else if (bottomLineCycler >= 1500) { // display model number and name
    if (eeprom::gd.modelNum < 10)
      lcd.print(F(" "));

    lcd.print(eeprom::gd.modelNum);
    lcd.print(F(" "));
    lcd.print(eeprom::md.modelName);
  }
  else // display current bottom line 'instruction'
    lcd.print(bottom);
}

boolean menuCheck() { // allow forward/back on encoder or up/down on joystick to cycle through main screen and menu items
  int16_t change = 0;
#ifdef ENCODER
  if (menu && encoder::pressed) { // if encoder pressed when displaying menu items
    if (!encoderEdit) encoderBeginEdit = true;
    encoderEdit = true;
  }
  if (!encoderEdit) {
    encoder::setMultiplier(1);
    encoderBeginEdit = false;
    change = encoder::change(change);
  }

  if (menu && !encoder::timeout) { // exit from menus back to main screens after thirty seconds of inactivity
    menu = 0;
    encoderEdit = false;
    encoder::setMultiplier(1);
    lcd.noBlink();
    return true;
  }
#else
  if (joystick::button & 0x04) { // down
    change = 1;
  } else if (joystick::button & 0x08) { // up
    change = -1;
  }

  if (menu && menu != 24 && joystick::button & 0x10) { // if joystick pressed when displaying menu items (except sevo sweep)
    menu = 0; // return to main screens
    return true;
  }
  if (menu && !joystick::timeout) { // exit from menus back to main screens after thirty seconds of inactivity
    menu = 0;
    lcd.noBlink();
    return true;
  }
#endif
  if (change > 0) {
    clearBottomLine();
    for (int8_t i = 0; i < change; i++) {
      if (++menu == 27) menu = 100; // skip gap
      if (!telemetry::modelHasTelemetry) { // skip telemetry pages if no telemetry
        if (menu >= 100 && menu != 255) menu = 0;
        if (menu == 22 || menu == 23) menu = 24;
      }
      if (menu == 103) menu = 0; // wrap 
      if (menu == 24 && mode != 0) menu = 25; // skip servo sweep if motor armed
#ifdef ENCODER
      if (menu == 25) menu = 26; // skip joystick mode in encoder version
#else
      if (menu == 1) menu = 2; // skip motor run time on joystick version as it's on the main screen
#endif
      if (menu >=2 && menu <= 4 && mode != 0) menu = 5; // skip load/rename/copy if motor armed
    }
  } else if (change < 0) {
    clearBottomLine();
    for (int8_t i = 0; i < -change; i++) {
      if (menu == 0) menu = 102; // wrap
      else if (--menu == 99) menu = 26; // skip gap
      if (menu >= 2 && menu <= 4  && mode != 0) menu = 1; // skip load/rename/copy if motor armed
#ifdef ENCODER
      if (menu == 25) menu = 24; // skip joystick mode in encoder version
#else
      if (menu == 1) menu = 0; // skip motor run time on joystick version as it's on the main screen
#endif
      if (menu == 24 && mode != 0) menu = 23; // skip servo sweep if motor armed
      if (!telemetry::modelHasTelemetry) { // skip telemetry pages if no telemetry
        if (menu >= 100  && menu != 255) menu = 26;
        if (menu == 22 || menu == 23) menu = 21;
      }
    }
  }
  return false;
}

void displayTempModel() {
  lcd.setCursor(0, 1);
  if (tempModelNum < 10)
    lcd.print(F(" "));

  lcd.print(tempModelNum);
  lcd.print(F(" "));
  lcd.print(eeprom::getModelName(tempModelNum));  
}

void selectTempModel() {
#ifdef ENCODER
  if (encoderEdit) {
    encoder::setMultiplier(1);
    int16_t change = 0;
    change = encoder::change(change);
    if (change || encoderBeginEdit) {
      tempModelNum += change;
      while (tempModelNum < 1) tempModelNum += 27;
      while (tempModelNum > 27) tempModelNum -= 27;
      lcd.noBlink();
      displayTempModel();
      lcd.setCursor(1, 1);
      lcd.blink();
    }
    if (!encoder::timeout) {
      lcd.noBlink();
      encoderEdit = false;
      encoder::setMultiplier(1);
      menu = 0;
      return;    
    }
  }
#else
  displayTempModel();

  if (joystick::button & 0x01) // right
    if (++tempModelNum > 27)
      tempModelNum = 1;
  if (joystick::button & 0x02) // left
    if (--tempModelNum < 1)
      tempModelNum = 27;
#endif
}

void edit_uS(const __FlashStringHelper *bottom, int16_t *uS)
{
#ifdef ENCODER
  int16_t change = 0;
  if (encoderEdit) {
    encoder::setMultiplier(10);    
    change = encoder::change(change);
    if (change) {
      *uS += change;
      if (*uS > 2200)
        *uS = 2200;
      if (*uS < 800)
        *uS = 800;
      eeprom::putMD(eeprom::gd.modelNum);
    }
    if (encoder::pressed) {
      encoderEdit = false;
    }
    if (!encoder::timeout) {
      lcd.noBlink();
      encoderEdit = false;
      encoder::setMultiplier(1);
      menu = 0;
      return;    
    }
  } else {
    lcd.noBlink();
    lcd.setCursor(0, 1);
    lcd.print(bottom); // should always be 9 print characters
    if (*uS < 1000)
      lcd.print(F(" "));
    lcd.print(*uS);
    lcd.print(F(" uS"));
    
    if (menuCheck()) {
      return;
    }
  }
#else
  if (menuCheck()) {
    return;
  }
  if (joystick::button & 0x01) { // right
    *uS += joystick::increment;
    if (*uS > 2200)
      *uS = 2200;
    eeprom::putMD(eeprom::gd.modelNum);
  }
  if (joystick::button & 0x02) { // left
    *uS -= joystick::increment;
    if (*uS < 800)
      *uS = 800;
    eeprom::putMD(eeprom::gd.modelNum);
  }
  if (!joystick::timeout) {
    menu = 0;
    return;    
  }

#endif

#ifdef ENCODER
  if (encoderEdit) {
    if (change) {
      lcd.noBlink();
      lcd.setCursor(0, 1);
      lcd.print(bottom); // should always be 9 print characters
      if (*uS < 1000)
        lcd.print(F(" "));
      lcd.print(*uS);
      lcd.print(F(" uS"));
    }
    lcd.setCursor(encoder::multiplying ? 11: 12, 1);
    lcd.blink();
  }
#else
  lcd.setCursor(0, 1);
  lcd.print(bottom); // should always be 9 print characters
  if (*uS < 1000)
    lcd.print(F(" "));
  lcd.print(*uS);
  lcd.print(F(" uS"));
#endif
}

void loop(void) {
  static uint32_t oldMillis = 0;
  uint32_t currentMillis = millis();
  uint16_t mS = currentMillis - oldMillis; // mS is milliseconds taken between consecutive calls to loop
  uint32_t buttonTime;

//  if (mS > 100)
//     mS = 0; // prevent startup delay causing problems
    
  oldMillis = currentMillis;

  if (mS) { // read battery voltage and smooth taking into account loop time
    batteryVolts *= 500 - mS;
    batteryVolts /= 500;
    batteryVolts += (float)eeprom::gd.batteryScale * analogRead(A2) * mS / 5120000;
  }

  if (second <= 500 && batteryVolts <= 0.1 * eeprom::gd.lowBatteryAlarm) {
    digitalWrite(8, HIGH); // buzzer on for battery warning
  }
  if (mode & 0x01 || mode == 2) { // if motor running, or paused in normal button mode - i.e. if timer is running
    if (remainingTime == 30 || remainingTime == 10 || remainingTime <= 5) { // 30-second, 10-second and last five second countdown bleeps
      digitalWrite(8, second > 500);
    }
    if (remainingTime == 1) { // last second - long bleep
      digitalWrite(8, HIGH);
    }
  }

  second -= mS;
  secondElapsed = false;
  if (second <= 0) {
    secondElapsed = true;
    second += 1000;
    digitalWrite(8, LOW); // buzzer off
  }

#ifdef ENCODER
  encoder::checkEncoder(mS);
  int16_t tens;
  int16_t change;
#else
  joystick::readJoy(mS);
#endif

  if (telemetry::modelHasTelemetry) {
    telemetry::check(mS);
  }

  switch(menu) {
    case 0: // main screens
      tempModelNum = eeprom::gd.modelNum; // load copy of current model number for menu item 1
      switch (mode) {
        case 0: // motor not running - allow left-right editing of motor run time in 10s increments
          displayTime(eeprom::md.motorRunTime);
#ifdef ENCODER
          refresh(F("Motor time "), F("Press dial start"), mS);
          if (encoder::pressed) { // encoder pressed - start motor
            remainingTime = eeprom::md.motorRunTime;
            mode = 1; 
          }          
#else
          refresh(F("Motor time "), F("Press joy starts"), mS);
          if (joystick::button & 0x01) { // right
            eeprom::md.motorRunTime += 10;
            if (eeprom::md.motorRunTime > 5990)
              eeprom::md.motorRunTime = 5990; // limit max to 99:50
            eeprom::putMD(eeprom::gd.modelNum);
          }
          if (joystick::button & 0x02) { // left
            if (eeprom::md.motorRunTime > 10)
              eeprom::md.motorRunTime -= 10;
            eeprom::putMD(eeprom::gd.modelNum);
          }
          if (joystick::button & 0x10) { // joystick pressed - start motor
            remainingTime = eeprom::md.motorRunTime;
            mode = 1; 
          }
#endif
          menuCheck();
          break;
        case 1: // motor running (normal button control)
#ifdef ENCODER
          refresh(F("Remaining  "), F("Press dial stops"), mS);
#else
          refresh(F("Remaining  "), F("Press joy stops "), mS);
#endif
          runTimer(secondElapsed);
          displayTimer();
#ifdef ENCODER
          if (encoder::pressed) // encoder pressed - stop motor
            mode = 0;
#else
          if (joystick::button & 0x10) // joystick pressed - stop motor
            mode = 0;
#endif
          menuCheck();
          break;
        case 2: // motor paused (quick blip)
#ifdef ENCODER
          refresh(F("Paused     "), F("Press dial stops"), mS);
#else
          refresh(F("Paused     "), F("Press joy stops "), mS);
#endif
          // runTimer(secondElapsed);
          displayTimer();
#ifdef ENCODER
          if (encoder::pressed) // encoder pressed - stop motor
            mode = 0;
#else
          if (joystick::button & 0x10) // joystick pressed - stop motor
            mode = 0;
#endif
          menuCheck();
          break;
#ifndef ENCODER          
        case 3: // joystick mode, motor running
          refresh(F("=Joystick= "), F("Turn off to exit"), mS);
          runTimer(secondElapsed);
          displayTimer();
          break;
        case 4: // joystick mode, motor paused
          refresh(F("=Joystick= "), F("Turn off to exit"), mS);
          displayTimer();
          break;
#endif
      }
      break;

    case 1: // motor run time (encoder version only)
#ifdef ENCODER
      if (encoderEdit) {
        encoder::setMultiplier(1);
        tens = 0;
        tens = encoder::change(tens) * 10;
        if (tens || encoderBeginEdit) {
          eeprom::md.motorRunTime += tens;
          if (eeprom::md.motorRunTime > 5990)
            eeprom::md.motorRunTime = 5990; // limit max to 99:50
          if (eeprom::md.motorRunTime < 10)
            eeprom::md.motorRunTime = 10; // limit min to 00:10
          eeprom::putMD(eeprom::gd.modelNum);
          lcd.noBlink();
          displayTime(eeprom::md.motorRunTime);
          lcd.setCursor(14, 1);
          lcd.blink();        
        }
        encoderBeginEdit = false;
        if (encoder::pressed || !encoder::timeout) {
          encoderEdit = false;
          encoder::setMultiplier(1);
          lcd.noBlink();
          menu = 0;
        }
      } else {
        displayTime(eeprom::md.motorRunTime);
        topLine(F("Motor time      "));
        displayTime(eeprom::md.motorRunTime);
        menuCheck();
      }
#endif      
      break;
    case 2: // select model
#ifdef ENCODER
      if (encoderEdit) {
        if (encoderBeginEdit) {
          topLine(F("Press to use    "));
          encoderBeginEdit = false;
          lcd.setCursor(1, 1);
          lcd.blink();
        }
        if (encoder::pressed) {
          eeprom::gd.modelNum = tempModelNum;
          eeprom::putGD();
          eeprom::getMD(eeprom::gd.modelNum);
          encoderEdit = false;
          lcd.noBlink();          
          mode = 0;
          menu = 0;
          telemetry::start(); // reset telemetry in case newly loaded model has none
        }
        if (!encoder::timeout) {
          encoderEdit = false;
          encoder::setMultiplier(1);
          lcd.noBlink();          
          menu = 0;          
        }
      } else {
        topLine(F("Press to select "));
        displayTempModel();
        menuCheck();
      }
      selectTempModel();
#else      
      topLine(F("Press to select "));
      selectTempModel();
      if (joystick::button & 0x10) { // pressed
        eeprom::gd.modelNum = tempModelNum;
        eeprom::putGD();
        eeprom::getMD(eeprom::gd.modelNum);
        mode = 0;
        menu = 0;
        telemetry::start(); // reset telemetry in case newly loaded model has none
      }
      menuCheck();
#endif
      break;
    case 3: // model name
      topLine(F("Press: edit name"));
      displayTempModel();
#ifdef ENCODER
      if (encoder::pressed) {
        posn = 0;
        lcd.setCursor(3 + posn, 1);
        lcd.blink();
        oldButtonState = 0;
        menu = 255; // special menu for editing model name
        break;
      }
#else
      if (joystick::button & 0x10) { // pressed
        posn = 0;
        lcd.setCursor(3 + posn, 1);
        lcd.blink();
        menu = 255; // special menu for editing model name
        break;
      }
#endif
      menuCheck();
      break;
    case 4: // copy model to
#ifdef ENCODER
      if (encoderEdit) {
        if (encoderBeginEdit) {
          topLine(F("Press to copy to"));
          encoderBeginEdit = false;
          lcd.setCursor(1, 1);
          lcd.blink();
        }
        if (encoder::pressed) {
          eeprom::gd.modelNum = tempModelNum;
          eeprom::putGD();
          eeprom::putMD(eeprom::gd.modelNum);
          encoderEdit = false;
          lcd.noBlink();          
          mode = 0;
          menu = 0;
        }
        if (!encoder::timeout) {
          encoderEdit = false;
          encoder::setMultiplier(1);
          lcd.noBlink();          
          mode = 0;
          menu = 0;
        }
      } else {
        topLine(F("Copy model to   "));
        displayTempModel();
        menuCheck();
      }
      selectTempModel();
#else
      topLine(F("Copy model to   "));
      selectTempModel();
      if (joystick::button & 0x10) { // pressed
        eeprom::gd.modelNum = tempModelNum;
        eeprom::putGD();
        eeprom::putMD(eeprom::gd.modelNum);
        mode = 0;
        menu = 0;
      }
#endif
      menuCheck();
      break;
    case 5: // rudder centre pulse time (Glide)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder centre  G"));      
#else
      topLine(F("Rudder centre  G"));
#endif
      edit_uS(F("         "), &(eeprom::md.rudderCentre[0]));
      break;
    case 6: // rudder centre pulse time (Power)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder centre  P"));      
#else
      topLine(F("Rudder centre  P"));
#endif
      edit_uS(F("         "), &(eeprom::md.rudderCentre[1]));
      break;
    case 7: // rudder right pulse time (Glide)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder 1-press G"));
#else
      topLine(F("Rudder 1-press G"));
#endif
      edit_uS(F("(Right)  "), &(eeprom::md.rudderRight[0]));
      break;
    case 8: // rudder right pulse time (Power)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder 1-press P"));
#else
      topLine(F("Rudder 1-press P"));
#endif
      edit_uS(F("(Right)  "), &(eeprom::md.rudderRight[1]));
      break;
    case 9: // rudder left pulse time (Glide)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder 2-press G"));
#else
      topLine(F("Rudder 2-press G"));
#endif
      edit_uS(F("(Left)   "), &(eeprom::md.rudderLeft[0]));
      break;
    case 10: // rudder left pulse time (Power)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder 2-press P"));
#else
      topLine(F("Rudder 2-press P"));
#endif
      edit_uS(F("(Left)   "), &(eeprom::md.rudderLeft[1]));
      break;
    case 11: // elevator centre pulse time (Glide)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elev centre    G"));
#else
      topLine(F("Elev centre    G"));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorCentre[0]));
      break;
    case 12: // elevator centre pulse time (Power)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elev centre    P"));
#else
      topLine(F("Elev centre    P"));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorCentre[1]));
      break;
    case 13: // elevator up pulse time (Glide)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elevator up    G"));
#else
      topLine(F("Elevator up    G"));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorUp[0]));
      break;
    case 14: // elevator up pulse time (Power)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elevator up    P"));
#else
      topLine(F("Elevator up    P"));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorUp[1]));
      break;
    case 15: // elevator down pulse time (Glide)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elevator down  G"));
#else
      topLine(F("Elevator down  G"));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorDown[0]));
      break;
    case 16: // elevator down pulse time (Power)
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elevator down  P"));
#else
      topLine(F("Elevator down  P"));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorDown[1]));
      break;
    case 17: // motor off pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Motor off       "));
#else
      topLine(F("Motor off       "));
#endif
      edit_uS(F("         "), &(eeprom::md.motorOff));
      break;
    case 18: // motor off pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Motor on        "));
#else
      topLine(F("Motor on        "));
#endif
      edit_uS(F("         "), &(eeprom::md.motorOn));
      break;
    case 19: // motor soft start time (milliseconds)
#ifdef ENCODER
      if (encoderEdit) {
        change = 0;
        change = encoder::change(change);
        if (change || encoderBeginEdit) {
          encoderBeginEdit = false;
          eeprom::md.motorSoftStart += change;          
          if (eeprom::md.motorSoftStart > 5000)
            eeprom::md.motorSoftStart = 5000;
          if (eeprom::md.motorSoftStart < 0)
            eeprom::md.motorSoftStart = 0;

          eeprom::putMD(eeprom::gd.modelNum);
          lcd.noBlink();
          lcd.setCursor(0, 1);
          lcd.print(F("         ")); // should always be 9 print characters
          if (eeprom::md.motorSoftStart < 1000)
            lcd.print(F(" "));
          if (eeprom::md.motorSoftStart < 100)
            lcd.print(F(" "));
          if (eeprom::md.motorSoftStart < 10)
            lcd.print(F(" "));
    
          lcd.print(eeprom::md.motorSoftStart);
          lcd.print(F(" mS"));
          lcd.setCursor(encoder::multiplying ? 11: 12, 1);
          lcd.blink();        
        }
        if (encoder::pressed) {
          lcd.noBlink();
          encoderEdit = false;
        }
        if (!encoder::timeout) {
          lcd.noBlink();
          encoderEdit = false;
          encoder::setMultiplier(1);
          menu = 0;
          return;    
        }
      } else {
        topLine(F("Motor soft start"));
        encoder::setMultiplier(10);
        if (menuCheck())
          break;

        lcd.setCursor(0, 1);
        lcd.print(F("         ")); // should always be 9 print characters
        if (eeprom::md.motorSoftStart < 1000)
          lcd.print(F(" "));
        if (eeprom::md.motorSoftStart < 100)
          lcd.print(F(" "));
        if (eeprom::md.motorSoftStart < 10)
          lcd.print(F(" "));
  
        lcd.print(eeprom::md.motorSoftStart);
        lcd.print(F(" mS"));        
      }
#else
      topLine(F("Motor soft start"));
      if (menuCheck())
        break;
      if (joystick::button & 0x01) { // right
        eeprom::md.motorSoftStart += joystick::increment;
        if (eeprom::md.motorSoftStart > 5000)
          eeprom::md.motorSoftStart = 5000;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      if (joystick::button & 0x02) { // left
        if (joystick::increment >= eeprom::md.motorSoftStart)
          eeprom::md.motorSoftStart = 0;
        else
          eeprom::md.motorSoftStart -= joystick::increment;

        eeprom::putMD(eeprom::gd.modelNum);
      }
      lcd.setCursor(0, 1);
      lcd.print(F("         ")); // should always be 9 print characters
      if (eeprom::md.motorSoftStart < 1000)
        lcd.print(F(" "));
      if (eeprom::md.motorSoftStart < 100)
        lcd.print(F(" "));
      if (eeprom::md.motorSoftStart < 10)
        lcd.print(F(" "));

      lcd.print(eeprom::md.motorSoftStart);
      lcd.print(F(" mS"));
#endif
      break;      
    case 20: // button speed (milliseconds)
#ifdef ENCODER
      if (encoderEdit) {
        change = 0;
        change = encoder::change(change);
        if (change || encoderBeginEdit) {
          encoderBeginEdit = false;
          eeprom::md.buttonSpeed += change;          
          if (eeprom::md.buttonSpeed > 1000)
            eeprom::md.buttonSpeed = 1000;
          if (eeprom::md.buttonSpeed < 100)
            eeprom::md.buttonSpeed = 100;

          eeprom::putMD(eeprom::gd.modelNum);
          lcd.noBlink();
          lcd.setCursor(0, 1);
          lcd.print(F("         ")); // should always be 9 print characters
          if (eeprom::md.buttonSpeed < 1000)
            lcd.print(F(" "));
    
          lcd.print(eeprom::md.buttonSpeed);
          lcd.print(F(" mS"));          
          lcd.setCursor(encoder::multiplying ? 11: 12, 1);
          lcd.blink();        
        }
        if (encoder::pressed) {
          lcd.noBlink();
          encoderEdit = false;
        }
        if (!encoder::timeout) {
          lcd.noBlink();
          encoderEdit = false;
          encoder::setMultiplier(1);
          menu = 0;
          return;    
        }
      } else {
        topLine(F("Button speed    "));
        encoder::setMultiplier(10);
        if (menuCheck())
          break;

        lcd.setCursor(0, 1);
        lcd.print(F("         ")); // should always be 9 print characters
        if (eeprom::md.buttonSpeed < 1000)
          lcd.print(F(" "));
  
        lcd.print(eeprom::md.buttonSpeed);
        lcd.print(F(" mS"));        
      }
#else
      topLine(F("Button speed    "));
      if (menuCheck())
        break;
      if (joystick::button & 0x01) { // right
        eeprom::md.buttonSpeed += joystick::increment;
        if (eeprom::md.buttonSpeed > 1000)
          eeprom::md.buttonSpeed = 1000;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      if (joystick::button & 0x02) { // left
        if (joystick::increment >= eeprom::md.buttonSpeed)
          eeprom::md.buttonSpeed = 100;
        else {
          eeprom::md.buttonSpeed -= joystick::increment;
          if (eeprom::md.buttonSpeed < 100)
            eeprom::md.buttonSpeed = 100;
        }
        eeprom::putMD(eeprom::gd.modelNum);
      }
      lcd.setCursor(0, 1);
      lcd.print(F("         ")); // should always be 9 print characters
      if (eeprom::md.buttonSpeed < 1000)
        lcd.print(F(" "));

      lcd.print(eeprom::md.buttonSpeed);
      lcd.print(F(" mS"));
#endif
      break;
    case 21: // motor control (quick blip) enable
      topLine(F("Motor control by"));
      if (menuCheck())
        break;
#ifdef ENCODER
      if (encoder::pressed) {
        if (eeprom::md.quickBlipEnabled)
          eeprom::md.quickBlipEnabled = 0;
        else
          eeprom::md.quickBlipEnabled = 1;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      encoderEdit = false;
#else        
      if (joystick::button & 0x03) { // left or right toggles true/false
        if (eeprom::md.quickBlipEnabled)
          eeprom::md.quickBlipEnabled = 0;
        else
          eeprom::md.quickBlipEnabled = 1;
        eeprom::putMD(eeprom::gd.modelNum);
      }
#endif
      lcd.setCursor(0, 1);
      lcd.print(F("quick blip?  "));
      if (eeprom::md.quickBlipEnabled)
        lcd.print(F("Yes"));
      else
        lcd.print(F("No "));
      
      break;
    case 22: // Model battery scaling
#ifdef ENCODER
      if (encoderEdit) {
        change = 0;
        change = encoder::change(change);
        if (change || encoderBeginEdit) {
          encoderBeginEdit = false;
          change += eeprom::md.Arange;          
          if (change > 255)
            change = 255;
          if (change < 0)
            change = 0;

          eeprom::md.Arange = change;

          eeprom::putMD(eeprom::gd.modelNum);
          lcd.noBlink();
          lcd.setCursor(0, 1);
          lcd.print(F("scaling    "));
          if (change < 100)
            lcd.print(F(" "));
    
          lcd.print(change / 10);
          lcd.print(F("."));
          lcd.print(change % 10);
          lcd.print(F("V"));
          lcd.setCursor(14, 1);
          lcd.blink();        
        }
        if (encoder::pressed) {
          lcd.noBlink();
          encoderEdit = false;
        }
        if (!encoder::timeout) {
          lcd.noBlink();
          encoderEdit = false;
          encoder::setMultiplier(1);
          menu = 0;
          return;    
        }
      } else {
        topLine(F("Model battery   "));
        encoder::setMultiplier(1);
        if (menuCheck())
          break;

        lcd.setCursor(0, 1);
        lcd.print(F("scaling    "));
        if (eeprom::md.Arange < 100)
          lcd.print(F(" "));
  
        lcd.print(eeprom::md.Arange / 10);
        lcd.print(F("."));
        lcd.print(eeprom::md.Arange % 10);
        lcd.print(F("V"));
      }
#else
      topLine(F("Model battery   "));
      if (menuCheck())
        break;
      if (joystick::button & 0x01) { // right
        if (255 - eeprom::md.Arange > joystick::increment)
          eeprom::md.Arange = 255;
        else
          eeprom::md.Arange += joystick::increment;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      if (joystick::button & 0x02) { // left
        if (joystick::increment >= eeprom::md.Arange)
          eeprom::md.Arange = 0;
        else 
          eeprom::md.Arange -= joystick::increment;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      lcd.setCursor(0, 1);
      lcd.print(F("scaling    "));
      if (eeprom::md.Arange < 100)
        lcd.print(F(" "));

      lcd.print(eeprom::md.Arange / 10);
      lcd.print(F("."));
      lcd.print(eeprom::md.Arange % 10);
      lcd.print(F("V"));
#endif
      break;
    case 23: // Model battery alarm
#ifdef ENCODER
      if (encoderEdit) {
        change = 0;
        change = encoder::change(change);
        if (change || encoderBeginEdit) {
          encoderBeginEdit = false;
          change += eeprom::md.Aalarm;          
          if (change > 255)
            change = 255;
          if (change < 0)
            change = 0;

          eeprom::md.Aalarm = change;

          eeprom::putMD(eeprom::gd.modelNum);
          lcd.noBlink();
          lcd.setCursor(0, 1);
          lcd.print(F("alarm      "));
          if (change < 100)
            lcd.print(F(" "));
    
          lcd.print(change / 10);
          lcd.print(F("."));
          lcd.print(change % 10);
          lcd.print(F("V"));
          lcd.setCursor(14, 1);
          lcd.blink();        
        }
        if (encoder::pressed) {
          lcd.noBlink();
          encoderEdit = false;
        }
        if (!encoder::timeout) {
          lcd.noBlink();
          encoderEdit = false;
          encoder::setMultiplier(1);
          menu = 0;
          return;    
        }
      } else {
        topLine(F("Model battery   "));
        encoder::setMultiplier(1);
        if (menuCheck())
          break;

        lcd.setCursor(0, 1);
        lcd.print(F("alarm      "));
        if (eeprom::md.Aalarm < 100)
          lcd.print(F(" "));
  
        lcd.print(eeprom::md.Aalarm / 10);
        lcd.print(F("."));
        lcd.print(eeprom::md.Aalarm % 10);
        lcd.print(F("V"));
      }
#else
      topLine(F("Model battery   "));
      if (menuCheck())
        break;
      if (joystick::button & 0x01) { // right
        if (255 - eeprom::md.Aalarm > joystick::increment)
          eeprom::md.Aalarm = 255;
        else
          eeprom::md.Aalarm += joystick::increment;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      if (joystick::button & 0x02) { // left
        if (joystick::increment >= eeprom::md.Aalarm)
          eeprom::md.Aalarm = 0;
        else 
          eeprom::md.Aalarm -= joystick::increment;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      lcd.setCursor(0, 1);
      lcd.print(F("alarm      "));
      if (eeprom::md.Aalarm < 100)
        lcd.print(F(" "));

      lcd.print(eeprom::md.Aalarm / 10);
      lcd.print(F("."));
      lcd.print(eeprom::md.Aalarm % 10);
      lcd.print(F("V"));
#endif
      break;
    case 24: // Servo sweep mode
      if (mode != 6) { // sweep mode not active
#ifdef ENCODER
        topLine(F("Press Enc starts"));
#else
        topLine(F("Press Joy starts"));
#endif
        lcd.setCursor(0, 1);
        lcd.print(F("servo sweep     "));
        if (menuCheck())
          break;
#ifdef ENCODER
        if (encoder::pressed) { // pressed
          topLine(F("Press Enc stops "));
          oldSweepMode = 0;
          sweepMode = 1;
          encoder::setMultiplier(1);
          mode = 6;
          break;
        }
#else
        if (joystick::button & 0x10) { // pressed
          topLine(F("Press Joy stops "));
          oldSweepMode = 0;
          sweepMode = 1;
          mode = 6;
          break;
        }
#endif
      } else { // mode == 6: sweep mode active
        if (oldSweepMode != sweepMode) {
          lcd.noBlink();
          lcd.setCursor(15, 1);
          lcd.print(sweepMode);
          lcd.setCursor(15, 1);
          lcd.blink();
          oldSweepMode = sweepMode;        
        }
#ifdef ENCODER
        change = sweepMode;
        change = encoder::change(change);
        if (change > 9) change = 9;
        if (change < 1) change = 1;
        sweepMode = change;
        if (encoder::pressed) {
          encoder::setMultiplier(1);          
          mode = 0;
          lcd.noBlink();
          encoderEdit = false;
        }
#else
        if (joystick::button & 0x01) { // right
          if (++sweepMode > 9)
            sweepMode = 9;
        }
        if (joystick::button & 0x02) { // left
          if (--sweepMode < 1)
            sweepMode = 1;
        }
        if (joystick::button & 0x10) { // pressed
          mode = 0;
          lcd.noBlink();
        }
#endif    
      }
      break;
#ifndef ENCODER
    case 25: // press for joystick mode
      topLine(F("Press joy starts"));
      lcd.setCursor(0, 1);
      lcd.print(F("joystick mode   "));
      
      if (joystick::button & 0x10) { // pressed
        remainingTime = eeprom::md.motorRunTime;        
        mode = 4;
        menu = 0;
      }
      menuCheck();
      break;
#endif
    case 26: // low battery alarm
#ifdef ENCODER
      if (encoderEdit) {
        change = 0;
        change = encoder::change(change);
        if (change || encoderBeginEdit) {
          encoderBeginEdit = false;
          eeprom::gd.lowBatteryAlarm += change;          
          if (eeprom::gd.lowBatteryAlarm > 160)
            eeprom::gd.lowBatteryAlarm = 160;
          if (eeprom::gd.lowBatteryAlarm < 65)
            eeprom::gd.lowBatteryAlarm = 65;

          eeprom::putGD();
          lcd.noBlink();
          lcd.setCursor(0, 1);
          lcd.print(F("alarm      "));
          if (eeprom::gd.lowBatteryAlarm < 100)
            lcd.print(F(" "));
    
          lcd.print(eeprom::gd.lowBatteryAlarm / 10);
          lcd.print(F("."));
          lcd.print(eeprom::gd.lowBatteryAlarm % 10);
          lcd.print(F("V"));
          lcd.setCursor(14, 1);
          lcd.blink();        
        }
        if (encoder::pressed) {
          lcd.noBlink();
          encoderEdit = false;
        }
        if (!encoder::timeout) {
          lcd.noBlink();
          encoderEdit = false;
          encoder::setMultiplier(1);
          menu = 0;
          return;    
        }
      } else {
        topLine(F("Low battery     "));
        encoder::setMultiplier(1);
        if (menuCheck())
          break;

        lcd.setCursor(0, 1);
        lcd.print(F("alarm      "));
        if (eeprom::gd.lowBatteryAlarm < 100)
          lcd.print(F(" "));
  
        lcd.print(eeprom::gd.lowBatteryAlarm / 10);
        lcd.print(F("."));
        lcd.print(eeprom::gd.lowBatteryAlarm % 10);
        lcd.print(F("V"));
      }
#else
      topLine(F("Low battery     "));
      if (menuCheck())
        break;
      if (joystick::button & 0x01) { // right
        eeprom::gd.lowBatteryAlarm += joystick::increment;
        if (eeprom::gd.lowBatteryAlarm > 160)
          eeprom::gd.lowBatteryAlarm = 160;
        eeprom::putGD();
      }
      if (joystick::button & 0x02) { // left
        if (joystick::increment >= eeprom::gd.lowBatteryAlarm)
          eeprom::gd.lowBatteryAlarm = 65;
        else {
          eeprom::gd.lowBatteryAlarm -= joystick::increment;
          if (eeprom::gd.lowBatteryAlarm < 65)
            eeprom::gd.lowBatteryAlarm = 65;
        }
        eeprom::putGD();
      }
      lcd.setCursor(0, 1);
      lcd.print(F("alarm      "));
      if (eeprom::gd.lowBatteryAlarm < 100)
        lcd.print(F(" "));

      lcd.print(eeprom::gd.lowBatteryAlarm / 10);
      lcd.print(F("."));
      lcd.print(eeprom::gd.lowBatteryAlarm % 10);
      lcd.print(F("V"));
#endif
      break;
    case 100: // telemetry RSSI
      topLine(F("Telemetry       "));
      lcd.setCursor(0, 1);
      lcd.print(F("RSSI         "));
      if (telemetry::rssi < 100)
        lcd.print(F(" "));
      if (telemetry::rssi < 10)
        lcd.print(F(" "));
      lcd.print(telemetry::rssi);
      menuCheck();
#ifdef ENCODER
      encoderEdit = encoderBeginEdit = false;
      encoder::setMultiplier(1);
#endif
      break;
    case 101: // telemetry A1
      topLine(F("Telemetry       "));
      lcd.setCursor(0, 1);
      lcd.print(F("A1         "));
      telemVolts = 3.3 * 4 * telemetry::a1 / 255;
      if (telemVolts < 10.0)
        lcd.print(F(" "));  
      lcd.print((int)telemVolts);
      lcd.print(F("."));
      telemVolts -= (int)telemVolts;
      lcd.print((int)(10 * telemVolts));
      lcd.print(F("V"));
      menuCheck();      
#ifdef ENCODER
      encoderEdit = encoderBeginEdit = false;
      encoder::setMultiplier(1);
#endif
      break;
    case 102: // telemetry A2
      topLine(F("Telemetry       "));
      lcd.setCursor(0, 1);
      lcd.print(F("A2         "));
      telemVolts = eeprom::md.Arange * telemetry::a2 / 2550;
      if (telemVolts < 10.0)
        lcd.print(F(" "));  
      lcd.print((int)telemVolts);
      lcd.print(F("."));
      telemVolts -= (int)telemVolts;
      lcd.print((int)(10 * telemVolts));
      lcd.print(F("V"));
      menuCheck();      
#ifdef ENCODER
      encoderEdit = encoderBeginEdit = false;
      encoder::setMultiplier(1);
#endif
      break;
    case 255: // editing model name
#ifdef ENCODER
    change = 0;
    change = encoder::change(change);
    buttonTime = button::milliseconds(&buttonUp);
    if (buttonUp) oldButtonState = 0;
    if (!buttonUp && oldButtonState == 0 && buttonTime > 20) { // real (non contact bounce) button press advances one character
      oldButtonState = 1;
      if (++posn > 12)
          posn = 0;
      change = 0;
    }
    if (!buttonUp && oldButtonState == 2 && buttonTime > 1000) { // long button press steps back one character
      oldButtonState = 3;
      posn -= 2;
      if (posn < 0)
          posn += 13;
      change = 0;
    }
    if (change) {
      char c = eeprom::modelName[posn] + change;
        while (c < 32)
          c += 95;
        while (c > 126)
          c -= 95;
      eeprom::modelName[posn] = c;
    }
    if (change || oldButtonState == 1 || oldButtonState == 3) {
      lcd.noBlink();
      lcd.setCursor(3, 1);
      lcd.print(eeprom::modelName);
      lcd.setCursor(3 + posn, 1);
      lcd.blink();
      if (oldButtonState == 1 || oldButtonState == 3) oldButtonState++;            
    }
    if (encoder::pressed) {
      for (posn = 0; posn < 13; posn++)
        eeprom::md.modelName[posn] = eeprom::modelName[posn];
      eeprom::putMD(eeprom::gd.modelNum);
      menu = 0;
      lcd.noBlink();
    }
    
    if (!encoder::timeout) { // inactivity timeout back to main screens (no save)
      menu = 0;
      lcd.noBlink();
      encoderEdit = false;
      encoder::setMultiplier(1);      
    }
#else    
      if (joystick::button & 0x01) // right
        if (++posn > 12)
          posn = 0;
      if (joystick::button & 0x02) // left
        if (!posn)
          posn = 12;
        else
          --posn;
      if (joystick::button & 0x04) { // down
        char c = eeprom::modelName[posn] - 1;
        if (c < 32)
          c = 126;
        eeprom::modelName[posn] = c;
      }
      if (joystick::button & 0x08) { // up
        char c = eeprom::modelName[posn] + 1;
        if (c > 126)
          c = 32;
        eeprom::modelName[posn] = c;
      }

      if (joystick::button) { // after any button press refresh
        lcd.noBlink();
        lcd.setCursor(3, 1);
        lcd.print(eeprom::modelName);
        lcd.setCursor(3 + posn, 1);
        lcd.blink();
      }
      
      if (joystick::button & 0x10) { // pressed
        for (posn = 0; posn < 13; posn++)
          eeprom::md.modelName[posn] = eeprom::modelName[posn];
        eeprom::putMD(eeprom::gd.modelNum);
        menu = 0;
        lcd.noBlink();
      }
      
      if (!joystick::timeout) { // inactivity timeout back to main screens (no save)
        menu = 0;
        lcd.noBlink();
      }
#endif
      break;
    default:
      lcd.setCursor(0, 1);
      lcd.print(menu);
      lcd.print(F(" = menu  "));
      break;
  }


  buttonTime = button::milliseconds(&buttonUp);

  if (mode == 3 || mode == 4) { // joystick mode (3 = motor running, 4 = motor paused)  button toggles motor
    if (buttonUp) {
      buttonMode = 0;
    } else if (!buttonMode && buttonTime > 20) { // each real (non contact bouncing) press toggles motor on/off
       buttonMode = 1;
       mode = 7 - mode;
      if (!remainingTime) {
        mode = 4; // can't restart motor once timer has run out
      }
    }
    // work rudder from joystick
#ifndef ENCODER      
    if (joystick::axis[0] > 20) { // right
      int32_t posn = (int32_t)eeprom::md.rudderRight[mode & 0x01] - (int32_t)eeprom::md.rudderCentre[mode & 0x01];
      posn *= (int32_t)joystick::axis[0];
      posn /= 512L;
      posn += (int32_t)eeprom::md.rudderCentre[mode & 0x01];
      cppm_out::set_uS(0, (int16_t)posn);
    } else if(joystick::axis[0] < -20) { // left
      int32_t posn = (int32_t)eeprom::md.rudderLeft[mode & 0x01] - (int32_t)eeprom::md.rudderCentre[mode & 0x01];
      posn *= (int32_t)joystick::axis[0];
      posn /= -512L;
      posn += (int32_t)eeprom::md.rudderCentre[mode & 0x01];
      cppm_out::set_uS(0, (int16_t)posn);
    } else { // rudder centre
      cppm_out::set_uS(0, eeprom::md.rudderCentre[mode & 0x01]);
    }

    // work elevator from joystick
    if (joystick::axis[1] > 20) { // joystick down -> up elevator
      int32_t posn = (int32_t)eeprom::md.elevatorUp[mode & 0x01] - (int32_t)eeprom::md.elevatorCentre[mode & 0x01];
      posn *= (int32_t)joystick::axis[1];
      posn /= 512L;
      posn += (int32_t)eeprom::md.elevatorCentre[mode & 0x01];
      cppm_out::set_uS(1, (int16_t)posn);
    } else if (joystick::axis[1] < -20) { // joystick up -> down elevator
      int32_t posn = (int32_t)eeprom::md.elevatorDown[mode & 0x01] - (int32_t)eeprom::md.elevatorCentre[mode & 0x01];
      posn *= (int32_t)joystick::axis[1];
      posn /= -512L;
      posn += (int32_t)eeprom::md.elevatorCentre[mode & 0x01];
      cppm_out::set_uS(1, (int16_t)posn);
    } else { // elevator centre
      cppm_out::set_uS(1, eeprom::md.elevatorCentre[mode & 0x01]);
    }
#endif
  } else if (mode == 6) { // servo sweep mode
    int16_t s1; // 'sawtooth l' - a bit like sine: range +511 ~ -512
    int16_t s2; // 'sawtooth 2' - a bit like cosine
    int16_t travel;
    int16_t limit;
    int16_t angle = (int16_t)(currentMillis & 2047UL); // cycles every 2.048 seconds
    if (angle < 512) {
      s1 = angle;
      s2 = 511 - angle;
    } else if (angle < 1024) {
      s1 = 1023 - angle;
      s2 = 511 - angle;
    } else if (angle < 1536) {
      s1 = 1023 - angle;
      s2 = angle - 1536;
    } else {
      s1 = angle - 2048;
      s2 = angle - 1536;
    }
/*    
    Serial.print(angle);
    Serial.print("\t");
    Serial.print(s1);
    Serial.print("\t");
    Serial.print(s2);
    Serial.print("\t");
 */    
    if (s1 >= 0) { // right rudder
      travel = eeprom::md.rudderRight[0] - eeprom::md.rudderCentre[0];
    } else { // left rudder
      travel = eeprom::md.rudderCentre[0] - eeprom::md.rudderLeft[0];      
    }
    if (travel >= 0) {
      limit = 2200 - eeprom::md.rudderCentre[0];
    } else {
      limit = 800 - eeprom::md.rudderCentre[0];
    }
    for (uint8_t i = 1; i < sweepMode; i++) {
      travel += travel >> 2;
      if (limit >= 0 && travel > limit) travel = limit;
      if (limit < 0 && travel < limit) travel = limit;
    }
    if (sweepMode == 9) { // in sweepMode 9, move servos full travel even if set travel is tiny
      travel = limit;
    }
    int32_t travel32 = (int32_t)travel;
    travel32 *= s1;
    travel32 >>= 9;
    travel = (int16_t)travel32;
    cppm_out::set_uS(0, eeprom::md.rudderCentre[0] + travel);

    if (s2 >= 0) { // up elevator
      travel = eeprom::md.elevatorUp[0] - eeprom::md.elevatorCentre[0];
    } else { // down elevator
      travel = eeprom::md.elevatorCentre[0] - eeprom::md.elevatorDown[0];      
    }
    if (travel >= 0) {
      limit = 2200 - eeprom::md.elevatorCentre[0];
    } else {
      limit = 800 - eeprom::md.elevatorCentre[0];
    }
    for (uint8_t i = 1; i < sweepMode; i++) {
      travel += travel >> 2;
      if (limit >= 0 && travel > limit) travel = limit;
      if (limit < 0 && travel < limit) travel = limit;
    }
    if (sweepMode == 9) { // in sweepMode 9, move servos full travel even if set travel is tiny
      travel = limit;
    }
    travel32 = (int32_t)travel;
    travel32 *= s2;
    travel32 >>= 9;
    travel = (int16_t)travel32;
    cppm_out::set_uS(1, eeprom::md.elevatorCentre[0] + travel);
  } else { // normal single channel (button) mode
    if (buttonUp) {
      if (buttonMode && buttonTime > eeprom::md.buttonSpeed * 2) { // when button released for long enough, back to idle mode (controls centred)
          buttonMode = 0;
      } else if (buttonTime > 20) { // a real release - not switch bounce
        switch (buttonMode) {
          case 1: // previously 'rotation started' - this is a quick blip so toggle motor state if enabled and return to idle
            if (eeprom::md.quickBlipEnabled) {
              if (mode == 1 || mode == 2) {
                mode = 3 - mode; // toggle motor state
                buttonMode = 0; // back to idle
              }
            } else {
              buttonMode = 4; // quick blip not enabled, so treat as a normal button press
            }
            break;
          case 2: // longer than a quick blip, but not 'RIGHT' 
            buttonMode = 4;
            break;
          case 3: // previously RIGHT
          case 6: // previously LEFT
          case 9: // previously KICK_UP
          case 11: // four presses but last press not long enough to kick down
          case 12: // previously KICK_DOWN
            buttonMode = 0; // back to idle
            break;
          case 5: // second release, (not 'LEFT')
            buttonMode = 7;
            break;
          case 8: // third release, (not 'UP')
            buttonMode = 10;
            break;
        }
      }
    } else if (buttonTime > 20) { // button really down - not switch bounce
      switch (buttonMode) {
        case 0: // idle - move to 'rotation started'
          buttonMode = 1;
          break;
        case 4:
          buttonMode = 5; // pressed 2nd time
          break;
        case 7:
          buttonMode = 8; // pressed 3rd time
          break;
        case 10:
          buttonMode = 11; // pressed 4th time
          break;
      }
      if (buttonTime > eeprom::md.buttonSpeed / 2) { // longer than a quick blip
        switch (buttonMode) {
          case 1:
            buttonMode = 2; // move to 'rotated past quick blip
            break;
        }
      }
      if (buttonTime > eeprom::md.buttonSpeed) { // button continued to be down long enough to move either rudder or elevator
        switch (buttonMode) {
          case 2: // single press and hold - RIGHT
            buttonMode = 3;
            break;
          case 5: // 1 press then press and hold - LEFT
            buttonMode = 6;
            break;
          case 8: // 2 presses then press and hold - KICK UP
            buttonMode = 9;
            break;  
          case 11: // 3 presses then press and hold - KICK DOWN
            buttonMode = 12;
            break;  
        }
      }
    }

    switch (buttonMode) {
      case 3:
        cppm_out::set_uS(0, eeprom::md.rudderRight[mode & 0x01]);
        break;
      case 6:
        cppm_out::set_uS(0, eeprom::md.rudderLeft[mode & 0x01]);
        break;
      case 9:
        cppm_out::set_uS(1, eeprom::md.elevatorUp[mode & 0x01]);
        break;
      case 12:
        cppm_out::set_uS(1, eeprom::md.elevatorDown[mode & 0x01]);
        break;
      default:
        cppm_out::set_uS(0, eeprom::md.rudderCentre[mode & 0x01]);
        cppm_out::set_uS(1, eeprom::md.elevatorCentre[mode & 0x01]);
      break;
    }
  }
  
  if (mode & 0x01) { // motor is running
    if (mS >= eeprom::md.motorSoftStart) {
      cppm_out::set_uS(2, eeprom::md.motorOn);
    } else {
      int16_t motorSpeed;
      if (eeprom::md.motorOn >= eeprom::md.motorOff) {
        motorSpeed = (eeprom::md.motorOn - eeprom::md.motorOff) * mS / eeprom::md.motorSoftStart; // change in speed this loop for soft start
        motorSpeed += cppm_out::get_uS(2);
      } else {
        motorSpeed = (eeprom::md.motorOff - eeprom::md.motorOn) * mS / eeprom::md.motorSoftStart; // change in speed this loop for soft start      
        motorSpeed = cppm_out::get_uS(2) - motorSpeed;
      }
      if (eeprom::md.motorOn > eeprom::md.motorOff && motorSpeed > eeprom::md.motorOn) motorSpeed = eeprom::md.motorOn;
      if (eeprom::md.motorOn < eeprom::md.motorOff && motorSpeed < eeprom::md.motorOn) motorSpeed = eeprom::md.motorOn;
      cppm_out::set_uS(2, motorSpeed);
    }
  } else { // motor stopped
    cppm_out::set_uS(2, eeprom::md.motorOff);    
  }
/*
  Serial.print(menu);
  Serial.print("\t");
  Serial.print(mode);
  Serial.print("\t");
  Serial.print(buttonUp);
  Serial.print("\t");
  Serial.print(buttonTime);
  Serial.print("\t");
  Serial.print(mS);
  Serial.print("\t");
  Serial.println(buttonMode);
 */
}
